﻿using System;
using System.IO;
using System.Threading;
using System.Windows.Forms;
using System.Diagnostics;
using System.Text;

namespace Alden_USB_CTR_15
{
    public partial class Form1 : Form
    {
        private System.Drawing.Image dotGreen, dotDark;
        public TextBox[] CountsDisplay;
        public PictureBox[] Counting;
        public ushort seconds;
        public Stream logStream;
        private DateTime logStartTime, logEndTime;
        private bool nonNumberEntered = false;
        private static Thread logThread = null;
        private double secondsPerStatusUpdate = 0.1;
        private int[] oldCounts;
        private int[] CountingTimeout = new int[15];
        bool logThreadDone = false;
        private int[] endCounts = new int[15];
        private uint RunNumber = 0;

        public Form1()
        {
            InitializeComponent();
            CountsDisplay = new TextBox[15];
            Counting = new PictureBox[15];
        }

        private void buttonLog_Click(object sender, EventArgs e)
        {
            logThreadDone = false;
            buttonLog.Enabled = false;
            AutoMonitorCheckbox.Visible = false;
            AbortButton.Visible = true;
            ushort.TryParse(logSeconds.Text, out seconds);

            RunNumber++;
            RunNumberLabel.Text = "Run: " + RunNumber.ToString();
            progressBar.Visible = true;

            timer1.Enabled = false;

            AIOUSBNet.AIOUSB.CTR_8254ModeLoad(AIOUSBNet.AIOUSB.diOnly, 0, (14) /* counter 14 */, 0, (ushort)(seconds - 1)); // seconds duration; counter 14 now outputs a desired duration
            AIOUSBNet.AIOUSB.CTR_8254Mode(AIOUSBNet.AIOUSB.diOnly, 0, 13, 0); // arm loaded count in 14 by manually issueing one clock out of 13
            AIOUSBNet.AIOUSB.CTR_8254Mode(AIOUSBNet.AIOUSB.diOnly, 0, 13, 1);

            oldCounts = null;
            for (uint i = 0; i < 12; i++)
            {
                var status = AIOUSBNet.AIOUSB.CTR_8254ModeLoad(AIOUSBNet.AIOUSB.diOnly, 0, i, 2, 0);
                if (status != 0) LogText.Text += "buttonLog status: " + status.ToString() + "\n";
            }

            logThread = new Thread(new ThreadStart(doLogThread))
            {
                Priority = ThreadPriority.Highest // J2H TODO: make config option
            };

            AIOUSBNet.AIOUSB.CTR_8254ModeLoad(AIOUSBNet.AIOUSB.diOnly, 0, (13) /* counter 13 */, 2, 10000); // divide-by-1000; counter 13 now outputs at 1Hz
            AIOUSBNet.AIOUSB.CTR_8254ModeLoad(AIOUSBNet.AIOUSB.diOnly, 0, (12) /* counter 12 */, 2, 1000); // divide-by-10000; counter 12 now outputs at 1kHz
                                                                                                           // output of (14) needs to be tied __through inverter__ to GATEs of counters (3-14)

            logThread.Start();
        }

        private void displayLogStatus(double progress)
        {
            progressBar.Value = Math.Min(Convert.ToInt32(progress * progressBar.Maximum), progressBar.Maximum);
            Tick();
        }

        private void doLogThread()
        {
            logStartTime = DateTime.Now;
            long beginTime = Stopwatch.GetTimestamp();
            long endTime = beginTime + seconds * Stopwatch.Frequency;
            long ticksPerStatusUpdate = Convert.ToInt64(Stopwatch.Frequency * secondsPerStatusUpdate);
            long nextStatusUpdateTime = beginTime + ticksPerStatusUpdate;
            byte[] buf = new byte[45];
            UInt32 buflen = 45;

            do
            {
                long now = Stopwatch.GetTimestamp();
                if ( now >= nextStatusUpdateTime )
                {
                    nextStatusUpdateTime += ticksPerStatusUpdate;
                    {
                        UInt16 v = 0;
                        byte s = 0xff;
                        AIOUSBNet.AIOUSB.CTR_8254ReadStatus(AIOUSBNet.AIOUSB.diOnly, 0, 14, ref v, ref s);
                        if ((s & 0x40) ==0) // if not NULL count
                            Invoke((MethodInvoker)delegate
                            {
                                displayLogStatus( (seconds-(v*1.0)) / seconds );
                            });

                        if ((s & 0x80) != 0) // OUT pin is HIGH
                            logThreadDone = true;
                         
                    }
                }
            } while (!logThreadDone);

            var status = AIOUSBNet.AIOUSB.GenericVendorRead(AIOUSBNet.AIOUSB.diOnly, 0x25, 0, 0, ref buflen, buf); // statusAll
            if (status != 0)
                LogText.Text += "Tick status: " + status.ToString() + "\n";
            if (buflen != 45)
                LogText.Text += "genread status len == " + buflen.ToString() + "\n";

            for (uint i = 0; i < 15; i++)
            {
                if ((buf[2 + i * 3] & 0b01000000) == 0) 
                    endCounts[i] = (65536 - (buf[i * 3 + 1] * 256 + buf[i * 3]) + 1) & 0xFFFF; // if not NULL COUNT
                else
                    endCounts[i] = 0;
            }
            logEndTime = DateTime.Now;

            Invoke((MethodInvoker)delegate {
                progressBar.Value = progressBar.Maximum;
                closeLog();
            });
        }

        private void closeLog()
        {
            // save endCounts to CSV, along with note for log row.text
            String LogLine = DateTime.Now.ToString("HH:mm:ss") + "," + RunNumber.ToString() + "," + seconds.ToString();
            for (int i = 0; i < 15; i++)
                LogLine += "," + endCounts[i].ToString()
            ;
            LogLine += ",\"" + NoteTextBox.Text + "\"";
            LogLine += "\r\n";
            byte[] LogBytes = new UTF8Encoding(true).GetBytes(LogLine);

            logStream = File.Open(logPath.Text, FileMode.Append, FileAccess.Write, FileShare.Read);
            logStream.Write(LogBytes, 0, LogBytes.Length);
            logStream.Close();

            TimingAccuracyLabel.Text = "Actual Duration: " + (logEndTime - logStartTime).ToString("g");

            timer1.Enabled = true;
            buttonLog.Enabled = true;
            AutoMonitorCheckbox.Visible = true;
            AbortButton.Visible = false;
            if (AutoMonitorCheckbox.Checked)
                AIOUSBNet.AIOUSB.CTR_8254Mode(AIOUSBNet.AIOUSB.diOnly, 0, 14, 0); // free-run GATE to counters when not logging
            progressBar.Visible = false;
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            init();
        }

        private void c11_CheckedChanged(object sender, EventArgs e)
        {
            (sender as CheckBox).Text = (sender as CheckBox).Checked ? "Counting" : "";
        }
        private void init()
        {
            var status = AIOUSBNet.AIOUSB.ResolveDeviceIndex(AIOUSBNet.AIOUSB.diOnly);
            if (status != 0) LogText.Text += "init status: " + status.ToString() + "\n";

            // J2H TODO Device Detection (support 2 cards!!) (EEPROMByte!)

            timer1.Enabled = false;
            dotGreen = c00.Image;
            dotDark = c01.Image;

            progressBar.Minimum = 0;
            progressBar.Maximum = 1000;

            for (uint i = 0; i < 15; i++)
            {
                Control[] edits = this.Controls.Find("textBox" + (i + 1).ToString(), true);
                edits[0].Text += " foo " + i.ToString();
                CountsDisplay[i] = edits[0] as TextBox;
                CountsDisplay[i].Text = "0";

                Control[] checks = this.Controls.Find("c" + i.ToString().PadLeft(2, '0'), true);
                Counting[i] = checks[0] as PictureBox;
                Counting[i].Image = dotDark;
                Counting[i].Tag = 0;

                CountingTimeout[i] = 0;
            }
            timer1.Enabled = true;

            { // J2H plausibly improvable
                logPath.Text = Environment.CurrentDirectory;
                if (logPath.Text[logPath.TextLength - 1] != '\\')
                    logPath.Text += '\\';
                logPath.Text += DateTime.Now.ToString("yyyyMMdd HHmmss") + " log.csv";
            }
            if (AutoMonitorCheckbox.Checked)
                AIOUSBNet.AIOUSB.CTR_8254Mode(AIOUSBNet.AIOUSB.diOnly, 0, 14, 0); // free-run GATE to counters when not logging

        }

        private void Tick()
        {
            int[] counts = new int[CountsDisplay.Length];

            byte[] buf = new byte[45];
            UInt32 buflen = 45;
            
            var status = AIOUSBNet.AIOUSB.GenericVendorRead(AIOUSBNet.AIOUSB.diOnly, 0x25, 0, 0, ref buflen, buf); //statusAll
            if (status != 0)
                LogText.Text += "Tick status: " + status.ToString() + "\n";
            if (buflen != 45)
                LogText.Text += "genread status len == " + buflen.ToString()+"\n";


            if (oldCounts == null)
            {
                oldCounts = counts;
                return;
            }

            for (uint i = 0; i < 15; i++)
            {
                if ((buf[2+i*3] & 0b01000000) == 0)
                    counts[i] = (65536 - (buf[i * 3 + 1] * 256 + buf[i * 3]) + 1) & 0xFFFF; // if not NULL COUNT
                else
                    counts[i] = 0;

                if (oldCounts[i] != counts[i])
                {
                    Counting[i].Image = dotGreen;
                    CountsDisplay[i].Text = counts[i].ToString();

                    CountingTimeout[i] = 10;
                }
                else if (CountingTimeout[i] < 0) { }
                else if (CountingTimeout[i] == 0)
                {
                    Counting[i].Image = dotDark;
                    CountingTimeout[i] = -1;
                }
                else
                {
                    CountingTimeout[i]--;
                }
            }
            oldCounts = counts;
        }


        private void timer1_Tick(object sender, EventArgs e)
        {
            Tick();
        }

        private void LogFileBrowse(object sender, EventArgs e)
        {
            saveFileDialog1.FileName = Path.GetFileName(logPath.Text);
            saveFileDialog1.InitialDirectory = Path.GetDirectoryName(logPath.Text);
            saveFileDialog1.Filter = "csv files (*.csv)|*.csv";
            saveFileDialog1.FilterIndex = 1;
            saveFileDialog1.RestoreDirectory = true;
            if ( saveFileDialog1.ShowDialog() == DialogResult.OK )
                logPath.Text = saveFileDialog1.FileName;
        }

        private void AutoMonitorCheckbox_CheckedChanged(object sender, EventArgs e)
        {
            if (AutoMonitorCheckbox.Checked)
                AIOUSBNet.AIOUSB.CTR_8254Mode(AIOUSBNet.AIOUSB.diOnly, 0, 14, 0); // free-run GATE to counters when not logging
            else
                AIOUSBNet.AIOUSB.CTR_8254Mode(AIOUSBNet.AIOUSB.diOnly, 0, 14, 1); // free-run GATE to counters when not logging
        }

        private void textBox17_KeyDown(object sender, KeyEventArgs e)
        {
            nonNumberEntered = false;
            if (e.KeyCode < Keys.D0 || e.KeyCode > Keys.D9)
                if (e.KeyCode < Keys.NumPad0 || e.KeyCode > Keys.NumPad9)
                    if (e.KeyCode != Keys.Back)
                        nonNumberEntered = true;

            if (Control.ModifierKeys == Keys.Shift)
                nonNumberEntered = true;
        }

        private void textBox17_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (nonNumberEntered == true)
                e.Handled = true;
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            logThreadDone = true;
            if (logThread != null)
                while (logThread.IsAlive)
                    Application.DoEvents();

        }

        private void AbortButton_Click(object sender, EventArgs e)
        {
            logThreadDone = true;
        }

    }
}
